#!/bin/sh

set -x

# Create links to temp. storage in /var, they do not exist
# if we update a pre-yocto system.
source /etc/init.d/rs-temp-dirs
source /sbin/update-functions.sh
source /etc/env

MIN_PART1_SIZE=65536
MIN_PART2_SIZE=2208
MIN_PART3_SIZE=131072
MIN_FLASH_SIZE=$(( $MIN_PART1_SIZE + $MIN_PART2_SIZE  + $MIN_PART3_SIZE ))

ROOTFS="rootfs.squashfs"
ROOTFS_DEBUG="rootfs.squashfs.debug"
OPTFS="optfs"
SUBFS="subfs"
DATAFS="varfs.tgz"
REMOVELIST="remove.list"
# modboot
BOOT="stage2.grub.img"
# set by /etc/env: KERNEL="$(upd_get_kernel_basename)"
RECOVERY="recovery.ext2fs"

if [ $(uname -r) = "3.4.20" ]; then 
	export LD_PRELOAD="/usr/lib/libvdso.so"
fi

if [ -z "${KERNEL}" ]; then
	KERNEL="bzImage"
fi

STEPS=3
UPDATE_TITLE="Update Step "
if [ ! -b "${1}" ] ; then
	ZIPFILE="${1}"
else
	ZIPFILE="/mnt/$2"
fi

if [ "$3" -eq "$3" ] 2>/dev/null ; then
	ZIP_OFFSET="$3"
else
	ZIP_OFFSET=0
fi

ZIPERR=/tmp/ziperr
CHKMSG="Checking update file..."
SAVE="/rw/etc.tar"

cat /proc/bus/pci/devices | cut -f2 | grep -i -q "80eebeef"
if [ $? -eq 0 ] ; then
	VIRTUALBOX=1
	IDLE=""
else
	VIRTUALBOX=0
	IDLE="idle=poll"
fi

FLASHDEV="hda"
if ! dd if=/dev/${FLASHDEV} of=/dev/null count=0 ; then
	FLASHDEV="sda"
fi

center()
{
echo -geometry ${1}x${2}+$(( ( ${xsize} - ${1} ) / 2))+$(( ( ${ysize} - ${2} ) / 2))
}

title()
{
echo "${UPDATE_TITLE}${1}/${STEPS}"
}

panic()
{
	qprogress_c -k 2>/dev/null
	qdialog $(center 240 100) -title "Update" ${qtoolargs} -c -1 "&Reboot" -o "$1"

	umount /rw
	umount /mnt

	reboot -f
}

panic_space()
{
	[ $# -gt 1 ] && rm -rf "$2"
	panic "write error or insufficent space on storage - $1! $2"
}

file_length()
{
stat -c '%s' ${1} 2>/dev/null || echo -n -1
}
 
check_ziperr()
{
if [ -s "${ZIPERR}" ] ; then
    ziperr="$(cat ${ZIPERR})"
    panic "Update file corrupted ($ziperr)!"
fi
}

zip_length()
{
echo " $(unzip -f ${ZIP_OFFSET} -l ${ZIPFILE} "${1}" 2>${ZIPERR} | grep "\<${1}$")" | tr -s ' ' | cut -f2 -d' '
check_ziperr
}

zip_read()
{
unzip -f ${ZIP_OFFSET} -p ${ZIPFILE} "${1}" 2>${ZIPERR}
check_ziperr
}

md5sum_cmp()
{
if [ ! -e "${1}" ] ; then
	MD5_ERR="checksum file missing"
	return 1
fi
if [ $(file_length "${1}") -ne ${3} ] ; then
	MD5_ERR="wrong file length"
	return 1
fi
if [ "$(md5sum "${1}" | tr -s ' ' | cut -f1 -d' ')" != "$(zip_read ${2}.md5sum)" ] ;then
	MD5_ERR="checksum verify failed"
	return 1
fi
return 0
}

md5sum_cmp_len()
{
[ ! -e ${1} ] && return 1
test "$(dd if=${1} count=1 bs=${3} 2>/dev/null | md5sum | tr -s ' ' | cut -f1 -d' ')" = "$(zip_read ${2}.md5sum)"
return $?
}

get_part_size()
{
SIZE="$(grep '\<'${FLASHDEV}${1}'$' /proc/partitions | tr -s ' ' | cut -f4 -d' ')"

if [ "${SIZE}" = "" ] ; then
	echo 0
else
	echo "${SIZE}"
fi
}

do_repart()
{
qdialog $(center 240 100) -title "Repartition" ${qtoolargs} -c -1 "&Reboot" -o "Repartioning not available..."

reboot -f
}

check_progress()
{
qprogress_c -s ${2} "${CHKMSG} ${1}"
}

check_progress_1()
{
echo -n x | check_progress "${1}" 1 >/dev/null
}

check_md5sum()
{
	local MD5SUM

	MD5SUM=$(zip_read ${1} | check_progress "${1}" "${2}" | md5sum | tr -s ' ' | cut -f1 -d' ')

	if [ "${MD5SUM}" != "$(zip_read ${1}.md5sum)" ] ; then
		panic "update image ${1} corrupted" 
	fi
}

check_cmp()
{
	local TMP_IMAGE=/tmp/image
	local MD5SUM

	if [ "${3}" != "" ] ; then
		MD5SUM=$(zip_read ${1} | check_progress "${1}" "${3}" | tee "${TMP_IMAGE}" | md5sum | tr -s ' ' | cut -f1 -d' ')

		if [ "${MD5SUM}" != "$(zip_read ${1}.md5sum)" ] ; then
			panic "update image ${1} corrupted" 
		fi

		check_progress "${1}" "${3}" <${TMP_IMAGE} | { cmp -s "${2}" ; ret=$? ; cat >/dev/null ; return $ret ; }

		ret=$?

		rm -f ${TMP_IMAGE}

		return $ret
	fi

	return 1
}

export HOME=/tmp
export DISPLAY=:0

if ! killall -0 udevd ; then
	/lib/udev/udevd --daemon
	/usr/bin/udevadm trigger --subsystem-match=input --action=add
fi

clear
clear >/dev/tty1
chvt 1

/usr/bin/X vt1 &
sleep 3

if ! killall -0 Xorg ; then
	if [ -f /etc/X11/xorg.conf.update ] ; then
		/usr/bin/Xorg -config /etc/X11/xorg.conf.update -br vt1 &
		sleep 3
	fi
fi
if ! killall -0 Xorg ; then
	if [ -f /tmp/xorg.conf ] ; then
		/usr/bin/Xorg -config /tmp/xorg.conf -br vt1 &
		sleep 3
	fi
fi
if ! killall -0 Xorg ; then
	if [ -f /tmp/XF86Config ] ; then
		/usr/bin/Xorg -config /tmp/XF86Config -br vt1 &
		sleep 3
	fi
fi
if ! killall -0 Xorg ; then
	/usr/bin/Xorg -config /etc/X11/xorg.conf -br vt1 &
	sleep 3
fi
if ! killall -0 Xorg ; then
	sed -e 's/"intel"/"fbdev"/' /etc/X11/xorg.conf >/tmp/xorg.conf
	/usr/bin/Xorg -config /tmp/xorg.conf -br vt1 &
	sleep 3
fi
# Try to start X server without a config as a last resort.
if ! killall -0 Xorg ; then
	/usr/bin/Xorg -config /dev/null -br vt1 &
	sleep 3
fi
if ! killall -0 Xorg ; then
    echo "Can't start x11, Xorg.0.log:"
    cat /var/log/Xorg.0.log
	reboot -f
fi

# Note: difference to previous buildsys4 images: use e16 instead of sWM:
if [ -x /usr/bin/sWM ]; then
    if [ -x /usr/share/sWM-env ] ; then
	    source /usr/share/sWM-env
    fi

    /usr/bin/sWM &
else
    # /usr/bin/e16 -t rs -f -Q /tmp &
    /usr/bin/e16 -f -Q /tmp &
    sleep 3
fi

xset s off
xset -dpms

if [ -b "$1" ] ; then
	if ! mount -o ro $1 /mnt ; then
		panic "unable to access update device" 
	fi
fi

cd /tmp

if ! grep -q  "\<${FLASHDEV}$" /proc/partitions ; then
	panic "no flash memory available" 
fi

FLASH_SIZE=$(get_part_size)
if [ $FLASH_SIZE -lt $MIN_FLASH_SIZE ] ; then
	panic "insufficent flash memory"
fi

PART1_SIZE=$(get_part_size 1)
PART2_SIZE=$(get_part_size 2)
PART3_SIZE=$(get_part_size 3)

if ! mount -o rw /dev/${FLASHDEV}3 /rw ; then
	panic "cannot access data partition"
fi

# free some storage space ...
rm -rf /rw/${ROOTFS}.new /rw/${OPTFS}.new
rm -rf /rw/etc.new ${SAVE} ${SAVE}.new
if [ "$(df /rw | tail -n 1 | tr -s ' ' | cut -f4 -d' ')" -lt 10240 ] ; then
	qdialog $(center 240 140) -title "$(title 3)" ${qtoolargs} -a -i -1 "&Reboot" -o "Software update failed.

At least 10 MB of free storage
space is requiered.

Please remove some user data."

	umount /mnt
	umount /rw

	reboot -f
fi

qprogress_c -k 2>/dev/null

ROOTFS_LEN=$(zip_length ${ROOTFS})
ROOTFS_DEBUG_LEN=$(zip_length ${ROOTFS_DEBUG})

RECOVERY_LEN=$(zip_length ${RECOVERY})
BOOT_LEN=$(zip_length ${BOOT})
KERNEL_LEN=$(zip_length ${KERNEL})
OPTFS_LEN=$(zip_length ${OPTFS})
DATAFS_LEN=$(zip_length ${DATAFS})
REMOVELIST_LEN=$(zip_length ${REMOVELIST})

if [ "${BOOT_LEN}" != "" ] ; then
	if [ $(( ${PART2_SIZE} * 1024 )) -lt ${BOOT_LEN} ] ; then
		panic "insufficent space on boot parition" 
	fi
	# KERNEL_LEN=""
fi

if [ "${RECOVERY_LEN}" != "" ] ; then
	if [ $(( ${PART1_SIZE} * 1024 )) -lt ${RECOVERY_LEN} ] ; then
		RECOVERY_LEN=""
	fi
fi

SUM=$((
	${BOOT_LEN:-0} +
	${KERNEL_LEN:-0} +
	${RECOVERY_LEN:-0} +
	${ROOTFS_DEBUG_LEN:-0} +
	${OPTFS_LEN:-0} +
	${DATAFS_LEN:-0} +
	${REMOVELIST_LEN:-0} +
	3
))

qprogress_s $(center 300 70) -title "$(title 1)" ${qtoolargs} -s ${SUM} "${CHKMSG}"

# modboot
if [ "${BOOT_LEN}" != "" ] ; then
	check_md5sum ${BOOT} ${BOOT_LEN}
	# if md5sum_cmp_len /dev/${FLASHDEV}2 ${BOOT} ${BOOT_LEN} ; then
		# BOOT_LEN=""
	# fi
fi

if [ "${KERNEL_LEN}" != "" ] ; then
	if md5sum_cmp /rw/${KERNEL} ${KERNEL} ; then
		KERNEL_LEN=""
	fi
fi

if [ "${RECOVERY_LEN}" != "" ] ; then
	check_md5sum ${RECOVERY} ${RECOVERY_LEN}
	if md5sum_cmp_len /dev/${FLASHDEV}1 ${RECOVERY} ${RECOVERY_LEN} ; then
		RECOVERY_LEN=""
	fi
fi

check_progress_1 ${ROOTFS}
if md5sum_cmp /rw/${ROOTFS} ${ROOTFS} ${ROOTFS_LEN} ; then
	ROOTFS_LEN=""
fi

[ "${OPTFS_LEN}" != "" ] && check_md5sum ${OPTFS} ${OPTFS_LEN}
[ "${DATAFS_LEN}" != "" ] && check_md5sum ${DATAFS} ${DATAFS_LEN}
[ "${REMOVELIST_LEN}" != "" ] && check_md5sum ${REMOVELIST} ${REMOVELIST_LEN}

check_progress_1 "Done"
qprogress_c -k 2>/dev/null

SUM=$((
	${BOOT_LEN:-0} +
	${KERNEL_LEN:-0} +
	${RECOVERY_LEN:-0} +
	${ROOTFS_LEN:-0} +
	${ROOTFS_DEBUG_LEN:-0} +
	${OPTFS_LEN:-0} +
	${DATAFS_LEN:-0} +
	${REMOVELIST_LEN:-0} +
	4
))

qprogress_s $(center 300 70) -title "$(title 2)" ${qtoolargs} -s ${SUM} "Performing update..."

mkdir /tmp/zcip
mv /rw/var/cache/zcip.* /tmp/zcip 2>/dev/null
rm -rf /rw/${SUBFS} /rw/devel /rw/testing /rw/opt /rw/startup /rw/var/cache/* 2>/dev/null
mv /tmp/zcip/zcip.* /rw/var/cache/ 2>/dev/null

grep -e "^${share_user}:" /rw/etc/passwd | head -n 1 >/rw/etc/passwd.user
tar -cf ${SAVE}.new -C /rw/etc \
 bluetooth.pin \
 rs232.conf \
 layout.conf \
 security/ \
 sysctl.conf \
 systemsecpw.srspw \
 vncpasswd \
 network/interfaces \
 samba/smbpasswd \
 samba/workgroup.conf \
 touch_cal \
 passwd.user \
 2>/dev/null
mv ${SAVE}.new ${SAVE}
sync

{
grep -e "^${share_user}:" -v /etc/passwd
{
grep -e "^${share_user}:" /rw/etc/passwd
grep -e "^${share_user}:" /etc/passwd
} | head -n 1
} >/tmp/passwd

mv /rw/etc /rw/etc.new
rm -rf /rw/etc.new

# modboot
if [ "${BOOT_LEN}" != "" ] ; then

    echo "Try to mount /dev/${FLASHDEV}2"

    #
    # Try to convert ext2 to ext3 since we have no ext2 filesystem 
    # built into the kernel.
    #

    tune2fs -j "/dev/${FLASHDEV}2"
    if ! mount "/dev/${FLASHDEV}2" /boot ; then
	if ! ( mkfs.ext3 -m 0 -L boot /dev/${FLASHDEV}2 && tune2fs -c0 -i0 "/dev/${FLASHDEV}2" && mount "/dev/${FLASHDEV}2" /boot ) ; then
		panic "cannot access boot partition"
	fi
    fi
    (
    zip_read ${BOOT}
    cat /dev/zero
    ) |
    qprogress_c -s ${BOOT_LEN} "Updating Stage2 Grub..." > /boot/bzImage
    # TODO: hardcoded grub.cfg here, generate from initial grub cfg on sda1....
    #
    cat <<EOF >"/boot/${BOOT}.cfg"
set timeout=0
set default=0
menuentry 'SecondStage Grub load (hd0,3)/bzImage' {
  set root='(hd0,3)'
  linux16   /bzImage rootdev=/dev/sda3 ide1=noprobe ide2=noprobe ide3=noprobe ide4=noprobe ide5=noprobe hdb=noprobe earlyprintk=null console=tty8 video=i810fb:mtrr,accel threadirqs quiet
}
EOF
    umount /boot
fi

if [ "${KERNEL_LEN}" != "" ] ; then
	(
	zip_read ${KERNEL}
	cat /dev/zero
	) |
	qprogress_c -s ${KERNEL_LEN} "Updating Kernel..." >/rw/${KERNEL}.new
	sync
	if ! md5sum_cmp /rw/${KERNEL}.new ${KERNEL} ${KERNEL_LEN} ; then
		panic_space "${MD5_ERR}" /rw/${KERNEL}.new
	fi
	chattr -i /rw/${KERNEL}
	mv -f /rw/${KERNEL}.new /rw/${KERNEL}
	chattr +i /rw/${KERNEL}
	sync
fi

if [ "${RECOVERY_LEN}" != "" ] ; then
	(
	zip_read ${RECOVERY}
	cat /dev/zero
	) |
	qprogress_c -s ${RECOVERY_LEN} "Updating Recovery..." >/dev/${FLASHDEV}1
fi

if [ "${REMOVELIST_LEN}" != "" ] ; then
	zip_read ${REMOVELIST} |
	qprogress_c -s ${REMOVELIST_LEN} "Remove obsolete files..." |
	while read l
	do
		eval rm -rf ${l}
	done
fi

rm -f /rw/${OPTFS} /rw/${OPTFS}.new

if [ "${ROOTFS_LEN}" != "" ] ; then
	rm -f /rw/${ROOTFS_DEBUG}
	cat /dev/ramdisk |
	qprogress_c -s ${ROOTFS_LEN} "Updating Rootfs..." >/rw/${ROOTFS}.new
	sync
	if ! md5sum_cmp /rw/${ROOTFS}.new ${ROOTFS} ${ROOTFS_LEN} ; then
		panic_space "${MD5_ERR}" /rw/${ROOTFS}.new
	fi
	mv -f /rw/${ROOTFS}.new /rw/${ROOTFS}
	sync
fi

if [ "${ROOTFS_DEBUG_LEN}" != "" ] ; then
	(
	zip_read ${ROOTFS_DEBUG}
	cat /dev/zero
	) |
	qprogress_c -s ${ROOTFS_DEBUG_LEN} "Debug Rootfs available, installing it..." >/rw/${ROOTFS_DEBUG}.new
	sync
	if ! md5sum_cmp /rw/${ROOTFS_DEBUG}.new ${ROOTFS_DEBUG} ${ROOTFS_DEBUG_LEN} ; then
		panic_space "${MD5_ERR}" /rw/${ROOTFS_DEBUG}.new
	fi
	mv -f /rw/${ROOTFS_DEBUG}.new /rw/${ROOTFS_DEBUG}
fi

(
echo -n 1
(
cd /rw
mount -o loop,ro -t squashfs /rw/${ROOTFS} /ro
cp -a /ro/etc etc.new || panic_space "failed to fetch /etc" /rw/etc.new
umount /ro
losetup -d /dev/loop0 2>/dev/null
if [ -s ${SAVE} ] ; then 
	tar -xf ${SAVE} -C etc.new || panic_space "failed to restore system config" /rw/etc.new
fi
cp /tmp/passwd etc.new/passwd || panic_space "failed to restore passwd" /rw/etc.new
mv etc.new etc
) >&2
sync
echo -n 2
) |
qprogress_c -s 2 "Update settings..." >/dev/null
rm -f ${SAVE} etc/passwd.user

if [ "${OPTFS_LEN}" != "" ] ; then
	(
	zip_read ${OPTFS}
	cat /dev/zero
	) |
	qprogress_c -s ${OPTFS_LEN} "Updating Application..." >/rw/${OPTFS}.new
	sync
	if ! md5sum_cmp /rw/${OPTFS}.new ${OPTFS} ${OPTFS_LEN} ; then
		panic_space "${MD5_ERR}" /rw/${OPTFS}.new
	fi
	mv -f /rw/${OPTFS}.new /rw/${OPTFS}
	sync
fi

if [ "${DATAFS_LEN}" != "" ] ; then
	(
	zip_read ${DATAFS}
	cat /dev/zero
	) |
	qprogress_c -s ${DATAFS_LEN} "Updating Application Data..." |
	tar -xz -C /rw
fi

if [ "${VIRTUALBOX}" -ne 0 ] ; then
	mkdir -p /rw/var/instrument/rc
	cat <<EOF >/rw/var/instrument/rc/99-simulation.ini
[Simulation]
SimulationMode=On
EOF
fi

(
echo -n 1
sync
sleep 2
echo -n 2
) |
qprogress_c -s 2 "Sync Filesystems..." >/dev/null

qprogress_c -k 2>/dev/null

umount /rw
umount /mnt

mount /dev/${FLASHDEV}1 /mnt
cat <<EOF >/mnt/boot/grub/grub.conf.new
timeout 0
default 0

title  System
 root (hd0,1)
 kernel /bzImage rootdev=/dev/sda3 ide1=noprobe ide2=noprobe ide3=noprobe ide4=noprobe ide5=noprobe hdb=noprobe earlyprintk=null console=tty8 video=i810fb:mtrr,accel threadirqs quiet

title  Recovery
 root (hd0,0)
 kernel /boot/bzImage root=/dev/hda1
EOF
mv /mnt/boot/grub/grub.conf.new /mnt/boot/grub/grub.conf
umount /mnt

if [ -b "${1}" ] ; then
qdialog $(center 240 140) -title "$(title 3)" ${qtoolargs} -m -a -i -1 "&Reboot" -o "Software update successful.

Please remove USB Stick
and press any key to
reboot instrument"
fi

reboot -f

